/*
Module : MfcCDDB.h
Purpose: Defines the interface for a collection of MFC classes to wrap access to CDDB

Copyright (c) 1999 - 2023 by PJ Naughter (Web: www.naughter.com, Email: pjna@naughter.com)

All rights reserved.

Copyright / Usage Details:

You are allowed to include the source code in any product (commercial, shareware, freeware or otherwise) 
when your product is released in binary form. You are allowed to modify the source code in any way you want 
except you cannot modify the copyright details at the top of each module. If you want to distribute source 
code with your application, then you are only allowed to distribute versions released by the author. This is 
to maintain a single distribution point for the source code. 

*/


//////////////////// Macros / Defines /////////////////////////////////////////

#pragma once

#ifndef __MFCCDDB_H__
#define __MFCCDDB_H__

#ifndef _EXCEPTION_
#pragma message("To avoid this message, you should put exception in your pre compiled header (normally stdafx.h)")
#include <exception>
#endif //#ifndef _EXCEPTION_
#ifndef _VECTOR_
#pragma message("To avoid this message, you should put exception in your pre compiled header (normally stdafx.h)")
#include <vector>
#endif //#ifndef _VECTOR_
#ifndef _STRING_
#pragma message("To avoid this message, you should put exception in your pre compiled header (normally stdafx.h)")
#include <string>
#endif //#ifndef _STRING_

#include "WinHTTPWrappers.h" //If you get a compilation error about this missing header file, then you need to download my WinHttpWrappers classes from http://www.naughter.com/winhttpwrappers.html

#ifndef MFCCDDB_EXT_CLASS
#define MFCCDDB_EXT_CLASS
#endif //#ifndef MFCCDDB_EXT_CLASS


//////////////////// Classes //////////////////////////////////////////////////

namespace CDDB
{

//Forward declarations
class CTrackPosition;
class CSite;
class CQueryResult;

//Typedefs
#ifdef _UNICODE
using String = std::wstring;
#else
using String = std::string;
#endif //#ifdef _UNICODE
using StringA = std::string;


class MFCCDDB_EXT_CLASS CException : public std::exception
{
public:
//Constructors / Destructors
  CException(_In_ HRESULT hr, _In_z_ LPCSTR pszLastResponse);
  CException(_In_ DWORD dwError, _In_ DWORD dwFacility, _In_z_ LPCSTR pszLastResponse);
  CException(_In_ bool bMMError, _In_ DWORD dwError, _In_z_ LPCSTR pszLastResponse);

//Methods
  bool GetErrorMessage(_Out_z_cap_(nMaxError) LPTSTR lpszError, _In_ UINT nMaxError);

//Data members
  bool m_bMMError;
  HRESULT m_hr;
  StringA m_sLastResponse;
};

//Class which encapsulates a CDDB site
class MFCCDDB_EXT_CLASS CSite
{
public:
//Constructors / Destructors
  CSite(_In_ bool bRetrieve = true);
  CSite(const CSite& site) = default;
  CSite(CSite&& site) = default;
  ~CSite() = default;

//Methods
  CSite& operator=(const CSite& site) = default;
  CSite& operator=(CSite&& site) = default;

//Data variables
  String m_sSite; //The site name e.g. "gnudb.gnudb.org"
  INTERNET_PORT m_nPort; //The port number to connect on, normally the HTTP port i.e. 80
  String m_sAddress; //The address of the HTTP file to use, normally either "/~cddb/cddb.cgi" or "/~cddb/submit.cgi"
  bool m_bNorthing; //Is this site north of the equator
  int m_nLatitudeMinutes; //Minutes of degrees north of the equator 
  bool m_bEasting; //Is this site east of the prime meridian
  int m_nLongitudeMinutes; //Minutes of degrees east of the prime meridian
  String m_sDescription; //Short description of the geographical location of the site
  String m_sProtocol; //The protocol the server uses (Note this is only read and not actively used by MfcCDDB)
};

//Class which represents the single instance of the results of a call to CCDDB::Query
class MFCCDDB_EXT_CLASS CQueryResult
{
public:
//Data variables
  String m_sCategory; //The category of this CD
  DWORD m_dwDiscID{0}; //The CDDB DISCID of this CD
  String m_sArtist; //The Artist of this CD
  String m_sTitle; //The Disc title of the CD
};

//Class which represents a single instance of the results as returned from CCDDB::Read and used in the call to CCDDB::Submit
class MFCCDDB_EXT_CLASS CRecord
{
public:
//Data variables
  std::vector<int> m_TrackOffsets; //The offsets of each track in frames
  int m_nDiskLength{0}; //The Disc length in seconds
  int m_nDatabaseRevision{0}; //The database revision number of this record
  String m_sClientName; //The name of the client which has submitted this record
  String m_sClientVersion; //The version of the client which has submitted this record
  String m_sClientComments; //Optional client comments
  DWORD m_dwDiscID{0}; //The CDDB DISC-ID of this CD
  String m_sArtist; //The Artist of this CD
  String m_sTitle; //The Disc title of the CD
  std::vector<String> m_TrackTitles; //All the track titles on this CD
  String m_ExtendedData; //The extended data for this CD
  std::vector<String> m_ExtendedTrackData; //The extended data for each track on this CD
  std::vector<int> m_PlayOrder; //The track play order if any for this CD
  int m_nYear{0}; //Year in which the CD was released
  String m_sGenre; //The genre of the disc in textual form
};

//Class which represents the result as returned from CCDDB::Status
class MFCCDDB_EXT_CLASS CStatus
{
public:
//Member variables
  int m_nCurrentProtocol{0}; //The servers current operating protocol level
  int m_nMaxProtocol{0}; //The maximum supported protocol level
  bool m_bGetsAllowed{false}; //Whether or not the client is allowed to get log information
  bool m_bUpdatesAllowed{false}; //Whether or not the client is allowed to initiate a database update
  bool m_bPostingAllowed{false}; //Whether or not the client is allowed to post new entries
  bool m_bQuotes{false}; //Whether or not quoted arguements are enabled
  int m_nCurrentUsers{0}; //The number of users currently connected to the server
  int m_nMaxUsers{0}; //The number of users that can concurrently connect to the server
  bool m_bStripExtended{false}; //Whether or not extended data is stripped by the server before presented to the user
  int m_nDatabaseEntries{0}; //Total number of entries in the database
  std::vector<String> m_Categories; //The categories held in the database
  std::vector<int> m_CategoryEntries; //The number of entries for each category in the database
  std::vector<String> m_PendingSites; //List of sites that are fed new database entries from this server
  std::vector<int> m_PendingEntries; //The number of entries that have yet to be transmitted to each site
};

//Simple class to store a tracks position in MSF format. This is required for computing the CDDB DISCID
class MFCCDDB_EXT_CLASS CTrackPosition
{
public:
//Member variables
  int m_nMinute{0};
  int m_nSecond{0};
  int m_nFrame{0};
};

//The main CDDB support class
class MFCCDDB_EXT_CLASS CClient
{
public:
//Constructors / Destructors
  CClient();
  CClient(const CClient&) = delete;
  CClient(CClient&&) = delete;
  virtual ~CClient() = default;

//Methods
  CClient& operator=(const CClient&) = delete;
  CClient& operator=(CClient&&) = delete;
  void GetCDROMDrives(_Out_ std::vector<String>& drives);
  DWORD ComputeDiscIDUsingMCI(_In_ LPCTSTR pszDrive);
  DWORD ComputeDiscIDUsingDeviceIoControl(_In_ LPCTSTR pszDrive);
  DWORD ComputeDiscID(_In_ const std::vector<CTrackPosition>& tracks);
  void GetTrackPositionsUsingMCI(_In_ LPCTSTR pszDrive, _Out_ std::vector<CTrackPosition>& tracks);
  void GetTrackPositionsUsingDeviceIoControl(_In_ LPCTSTR pszDrive, _Out_ std::vector<CTrackPosition>& tracks);
  void AttachSession(_In_ HINTERNET hSession);
  HINTERNET DetachSession();
  void SetCDDBProtocolVersion(_In_ int nProtocolVersion) noexcept;
  _NODISCARD int GetCDDBProtocolVersion() const noexcept { return m_nProtocolVersion; };
  void Sites(_Out_ std::vector<CSite>& sites, _In_ const String& = _T("gnudb.gnudb.org"), _In_ const String& sAddress = _T("/~cddb/cddb.cgi"), _In_ INTERNET_PORT nPort = 80);
  void Sites(_In_ const CSite& server, _Out_ std::vector<CSite>& sites);
  void Categories(_In_ const CSite& server, _Out_ std::vector<String>& categories);
  void Status(_In_ const CSite& server, _Out_ CStatus& status);
  void Query(_In_ const CSite& server, _In_ DWORD dwDiscID, _In_ const std::vector<CTrackPosition>& tracks, _Out_ std::vector<CQueryResult>& results);
  void Read(_In_ const CSite& server, _In_ DWORD dwDiscID, _In_ const String& sCategory, _Out_ CRecord& record);
  void Submit(_In_ const CSite& server, _In_ const String& sCategory, _In_ const String& sEmailAddress, _In_ const CRecord& record, _In_ const String& sRejectionNotice, _In_ bool bReallySubmit = true);
  void MessageOfTheDay(_Inout_ String& sMessage, _In_ const String& sServer = _T("gnudb.gnudb.org"), _In_ const String& sAddress = _T("/~cddb/cddb.cgi"), _In_ INTERNET_PORT nPort = 80);
  void MessageOfTheDay(_In_ const CSite& server, _Inout_ String& sMessage);
  _NODISCARD String GetProductName() const { return m_sProductName; };
  void SetProductName(_In_ const String& sProductName) { m_sProductName = sProductName; };
  _NODISCARD std::wstring GetUserAgent() const { return m_sUserAgent; };
  void SetUserAgent(_In_ const std::wstring& sUserAgent) { m_sUserAgent = sUserAgent; };
  _NODISCARD String GetProductVersion() const { return m_sProductVersion; };
  void SetProductVersion(const String& sProductVersion) { m_sProductVersion = sProductVersion; };
  _NODISCARD String GetHostName() { return m_sHostName; };
  void SetHostname(_In_ const String& sHostname);
  _NODISCARD String GetUserName() { return m_sUserName; };
  void SetUserName(_In_ const String& sUserName);
  virtual String GetHelloCommand();
  virtual void PrepareTextForSubmit(_Inout_ String& sText);
  void StripNonAlphaOrDigits(_Inout_ String& sValue);
  virtual int GetCDDBResponseCode(_In_ LPCSTR pszBody);
  _Return_type_success_(return != 0) virtual bool SplitSemiColonLine(_In_ const StringA& sLine, _Out_ StringA& sField, _Out_ StringA& sValue);
  _Return_type_success_(return != 0) virtual bool ParseSitesBody(_In_ LPCSTR pszBody, _Out_ std::vector<CSite>& sites);
  _Return_type_success_(return != 0) virtual bool ParseSiteBodyHelper(_In_ const StringA& sLine, _Out_ CSite& site);
  _Return_type_success_(return != 0) virtual bool ParseCategoriesBody(_In_ LPCSTR pszBody, _Out_ std::vector<String>& categories);
  _Return_type_success_(return != 0) virtual bool ParseStatusBody(_In_ LPCSTR pszBody, _Out_ CStatus& status);
  _Return_type_success_(return != 0) virtual bool ParseQueryBody(_In_ LPCSTR pszBody, _Inout_ std::vector<CQueryResult>& results);
  _Return_type_success_(return != 0) virtual bool ParseQueryBodyHelper(_In_ const StringA& sLine, _Out_ CQueryResult& result);
  _Return_type_success_(return != 0) virtual bool ParseRecordBody(_In_ LPCSTR pszBody, _Out_ CRecord& record);
  _Return_type_success_(return != 0) virtual bool ParseRecordBodyHelper(_In_ const StringA& sRecord, _Out_ CRecord& record);
  _Return_type_success_(return != 0) virtual bool ParseMessageBody(_In_ LPCSTR pszBody, _Inout_ String& sMessage);
  _Return_type_success_(return != 0) virtual bool ParseSubmitBody(_In_ LPCSTR pszBody);
  virtual String FormSubmitLine(_In_ const String& sKeyword, _In_ const String& sValue);
  virtual String MapFromProtocolString(_In_ const StringA& sInput);
  virtual StringA MapToProtocolString(_In_ const String& sInput);
  virtual void GetNextLine(_Inout_ LPCSTR& pszLine, _Inout_ StringA& sLine);
  _Return_type_success_(return != 0) virtual bool ParseCommentLine(_In_ const StringA& sLine, _Out_ StringA& sValue);
  _Return_type_success_(return != 0) virtual bool ParseKeywordLine(_In_ const StringA& sLine, _Out_ StringA& sKeyword, _Out_ StringA& sValue);

//Static methods
  static void ThrowCDDBException(_In_ DWORD dwError = 0, _In_ DWORD Facility = FACILITY_WIN32, _In_z_ LPCSTR pszLastResponse = "");
  static void ThrowCDDBException(_In_ HRESULT hr, _In_z_ LPCSTR pszLastResponse = "");
  static void ThrowCDDBMMException(_In_ DWORD dwError, _In_z_ LPCSTR pszLastResponse = "");

protected:
#ifdef _UNICODE
  void StringReplaceAll(_Inout_ String& sString, _In_ const String& sFrom, _In_ const String& sTo);
  void StringTrim(_Inout_ String& sString);
#endif //#ifdef _UNICODE
  void StringReplaceAll(_Inout_ StringA& sString, _In_ const StringA& sFrom, _In_ const StringA& sTo);
  void StringTrim(_Inout_ StringA& sString);
  virtual HRESULT CreateSession();
  virtual HRESULT CreateConnection(_In_z_ LPCTSTR pszServer, _In_ INTERNET_PORT nPort);
  virtual HRESULT CreateRequest(_In_z_ LPCTSTR pszRequest, _In_opt_z_ LPCWSTR pwszVerb);
  virtual HRESULT ReadResponse(_In_reads_bytes_opt_(dwRequestSize) LPVOID pRequest, _In_ DWORD dwRequestSize, _In_ const WinHTTPWrappers::String& sHeaders, _Inout_ StringA& sResponse);

//Member variables
  String m_sProductName;
  String m_sProductVersion;
  int m_nProtocolVersion;
  String m_sHostName;
  String m_sUserName;
  std::wstring m_sUserAgent;
  WinHTTPWrappers::CSession m_Session;
  WinHTTPWrappers::CConnection m_Connection;
  WinHTTPWrappers::CSyncDownloader m_Request;
};


}; //namespace CDDB


#endif //#ifndef __MFCCDDB_H__
